from datetime import datetime as dt, timedelta
from tap_tester import runner, menagerie, connections
from base import IntercomBaseTest

class IntercomParentChildSync(IntercomBaseTest):
    @staticmethod
    def name():
        return "tap_tester_intercom_parent_child_sync"

    def test_run(self):
        # Run with parent stream as earlier bookmark
        self.run_test(
            child_bookmark=(dt.now()).strftime(self.START_DATE_FORMAT),
            parent_bookmark=(dt.now()-timedelta(days=2)).strftime(self.START_DATE_FORMAT),
            earlier_stream="conversations"
        )

        # Run with child stream as earlier bookmark
        self.run_test(
            child_bookmark=(dt.now()-timedelta(days=2)).strftime(self.START_DATE_FORMAT),
            parent_bookmark=(dt.now()).strftime(self.START_DATE_FORMAT),
            earlier_stream="conversation_parts"
        )

    def run_test(self, child_bookmark, parent_bookmark, earlier_stream):
        """
        Test case to verify the working of parent-child streams
        Prerequisite:
            - Set child bookmark is earlier than parent bookmark
            - Set Parent bookmark is earlier than child bookmark
        Assertions:
            - Verify the we got records from the set bookmark for the streams
            - Verify we sync some records for the earlier stream which is between the 
                both the bookmark dates ie. child stream bookmark and parent streams bookmark
        """
        expected_streams = {"conversations", "conversation_parts"}

        # Create state file
        state = {
            "bookmarks": {
                "conversation_parts": {
                    "updated_at": child_bookmark
                },
                "conversations": {
                    "updated_at": parent_bookmark
                }
            }
        }

        conn_id = connections.ensure_connection(self)

        # Run in check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # Select only the expected streams tables
        catalog_entries = [ce for ce in found_catalogs if ce['tap_stream_id'] in expected_streams]
        self.perform_and_verify_table_and_field_selection(conn_id, catalog_entries, select_all_fields=True)

        # Set state
        menagerie.set_state(conn_id, state)

        # Run a sync job using orchestrator
        sync_record_count = self.run_and_verify_sync(conn_id)
        sync_records = runner.get_records_from_target_output()
        bookmark = menagerie.get_state(conn_id)

        expected_replication_keys = self.expected_replication_keys()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # Collect information for assertions from the sync based on expected values
                records = [record.get('data') for record in sync_records.get(stream, {}).get('messages',{})
                           if record.get('action') == 'upsert']

                # Collect information specific to incremental streams from sync
                replication_key = next(iter(expected_replication_keys[stream]))
                simulated_bookmark_value = self.convert_state_to_utc(state['bookmarks'][stream][replication_key])
                stream_bookmark_from_state = self.convert_state_to_utc(bookmark.get('bookmarks', {stream: None}).get(stream).get(replication_key))

                # Verify the sync sets a bookmark of the expected form
                self.assertIsNotNone(stream_bookmark_from_state)

                # Verify we replicated some records for the stream
                self.assertGreater(sync_record_count.get(stream, 0), 0)

                for record in records:
                    # This is child stream and bookmark is being written of parent stream. Thus, skipping the stream for assertion
                    if stream != "conversation_parts":
                        # Verify the we got records from the set bookmark for the streams
                        replication_key_value = record.get(replication_key)
                        self.assertGreaterEqual(
                            replication_key_value, simulated_bookmark_value,
                            msg="Second sync records do not respect the previous bookmark.")

                # Verify we sync some records for the earlier stream which is between the both the
                # bookmark dates ie. child stream bookmark and parent streams bookmark
                if stream == earlier_stream:
                    records_between_dates = []
                    for record in records:
                        if self.convert_state_to_utc(parent_bookmark) < record.get(replication_key) < self.convert_state_to_utc(child_bookmark):
                            records_between_dates.append(record)

                    self.assertIsNotNone(records_between_dates)
